
/* 
 * File        : compression.c
 * Lecture     : Data Processing on Modern Hardware
 * Assignment  : 4	
 * Description : Compression in Column Stores (4.3)
 * Author      : Jens Teubner, Louis Woods, ETH Zurich, 2010-10-30
 * ---------------------------------------------------------------------
 * Changelog   : 20.11.2010 - (lwoods) modified to work with shell script 
 *             : 20.11.2010 - (lwoods) added sumup examples
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>

#include <sys/time.h>

#include <xmmintrin.h>
#include <tmmintrin.h>
#include <smmintrin.h>

/* Note : use a multiple of eight */
//#define N 1048576
//#define N  8388608
//#define N 1048576
#define N (1 << 25)

#ifndef MODE
	#define MODE 0
#endif

#ifndef SIMD
	#define SIMD 0
#endif

#ifndef SUMUP
	#define SUMUP 0
#endif

static void
usage (const char *progname)
{
    printf ("usage: %s -n [num]\n", progname);
    exit (EXIT_SUCCESS);
}

static unsigned int *
generate (unsigned long n)
{
    unsigned int *ret = malloc (n * sizeof (*ret));

    if (!ret)
    {
        fprintf (stderr, "error during memory allocation\n");
        exit (EXIT_FAILURE);
    }

    for (unsigned long i = 0; i < n; i++) {
#if ( MODE == 1 )
      ret[i] = random () % (1 << 7);
#elif ( MODE == 2 )
      ret[i] = random () % (1 << 9);
#else
      ret[i] = random () % (1 << 8);
#endif
    }

    return ret;
}

#if ( MODE == 1 )

static uint8_t *
compress32to7 (unsigned int *data, unsigned long n)
{

    unsigned long m = (n * 7L/8L) + 1;
    uint8_t *ret = malloc (m * sizeof (*ret));

    long long offset = 0;
    for (long long i_in = 0; i_in < n; i_in+=8)
    {
      ret[offset+0]  = data[i_in+0] & 0x7f;
      ret[offset+0] |= (data[i_in+1] & 0x1) << 7;

      ret[offset+1]  = (data[i_in+1] & 0x7f) >> 1;
      ret[offset+1] |= (data[i_in+2] & 0x3) << 6;

      ret[offset+2]  = (data[i_in+2] & 0x7f) >> 2;
      ret[offset+2] |= (data[i_in+3] & 0x7) << 5;

      ret[offset+3]  = (data[i_in+3] & 0x7f) >> 3;
      ret[offset+3] |= (data[i_in+4] & 0xf) << 4;

      ret[offset+4]  = (data[i_in+4] & 0x7f) >> 4;
      ret[offset+4] |= (data[i_in+5] & 0x1f) << 3;

      ret[offset+5]  = (data[i_in+5] & 0x7f) >> 5;
      ret[offset+5] |= (data[i_in+6] & 0x3f) << 2;

      ret[offset+6]  = (data[i_in+6] & 0x7f) >> 6;
      ret[offset+6] |= (data[i_in+7] & 0x7f) << 1;

      offset += 7;
    }

    return ret;
}

#if ( SIMD == 1 )
#if ( SUMUP == 1 )
static void
SIMD_decompress7to32_sum(long unsigned *sum, uint8_t *src, unsigned long n)
{
    uint8_t *end = src + n * 7 / 8;

    __m128i parallel_sum = _mm_setzero_si128();

    __m128i  ba;
    __m128i  ca;

     /* Note: Order of the arguments of _mm_set_epi8 is w15, w14, ..., w0! */
    const __m128i shufmask = _mm_set_epi8 (0xff, 0xff, 3, 2,
                                           0xff, 0xff, 2, 1,
                                           0xff, 0xff, 1, 0,
                                           0xff, 0xff, 1, 0);

    const __m128i shufmask2 = _mm_set_epi8 (0xff, 0xff, 3, 2,
                                           0xff, 0xff, 3, 2,
                                           0xff, 0xff, 2, 1,
                                           0xff, 0xff, 1, 0);

    const __m128i mul = _mm_set_epi32 (1 << ((32 - 7) - 5),
                                       1 << ((32 - 7) - 6),
                                       1 << ((32 - 7) - 7),
                                       1 << ((32 - 7) - 0));

    const __m128i mul2 = _mm_set_epi32 (1 << ((32 - 7) - 2 - 7),
                                       1 << ((32 - 7) - 2),
                                       1 << ((32 - 7) - 3),
                                       1 << ((32 - 7) - 4));

    while (src < end)
    {
        /** load values 0-15 **/
        ba = _mm_loadu_si128 ((__m128i *) src);


        /** (1) decompress values 0-3 **/

        /* shuffle */
        ca = _mm_shuffle_epi8 (ba, shufmask);

        /* use multiplication (pmulld) to shift to left */
        ca = _mm_mullo_epi32 (ca, mul);

        /* use psrld to shift right */
        ca = _mm_srli_epi32 (ca, (32 -7));

        /* add up result */
        parallel_sum = _mm_add_epi32(parallel_sum, ca);


        /** (2) decompress values 4-7 **/

        /* shift raw data by 4 bytes */
        ba = _mm_srli_si128 (ba, 3);

        /* shuffle */
        ca = _mm_shuffle_epi8 (ba, shufmask2);

        /* use multiplication (pmulld) to shift to left */
        ca = _mm_mullo_epi32 (ca, mul2);

        /* use psrld to shift right */
        ca = _mm_srli_epi32 (ca, (32 - 7));

        /* add up result */
        parallel_sum = _mm_add_epi32(parallel_sum, ca);


        /** (3) decompress values 8-11 **/

        /* shift raw data by 5 bytes */
        ba = _mm_srli_si128 (ba, 4);

        /* shuffle */
        ca = _mm_shuffle_epi8 (ba, shufmask);

        /* use multiplication (pmulld) to shift to left */
        ca = _mm_mullo_epi32 (ca, mul);

        /* use psrld to shift right */
        ca = _mm_srli_epi32 (ca, (32 - 7));

        /* add up result */
        parallel_sum = _mm_add_epi32(parallel_sum, ca);


        /** (4) decompress values 12-15 **/

        /* shift raw data by 4 bytes */
        ba = _mm_srli_si128 (ba, 3);

        /* shuffle */
        ca = _mm_shuffle_epi8 (ba, shufmask2);

        /* use multiplication (pmulld) to shift to left */
        ca = _mm_mullo_epi32 (ca, mul2);

        /* use psrld to shift right */
        ca = _mm_srli_epi32 (ca, (32 - 7));

        /* add up result */
        parallel_sum = _mm_add_epi32(parallel_sum, ca);

        /** 16x7bit = 14 byte **/
        src += 14;
    }

    *sum = _mm_extract_epi32(parallel_sum,0)+_mm_extract_epi32(parallel_sum,1)+_mm_extract_epi32(parallel_sum,2)+_mm_extract_epi32(parallel_sum,3);
}
#else
static void
SIMD_decompress7to32 (unsigned int *dest, uint8_t *src, unsigned long n)
{
    uint8_t *end = src + n * 7 / 8;

    __m128i *out = (__m128i *) dest;

    __m128i  ba;
    __m128i  ca;

     /* Note: Order of the arguments of _mm_set_epi8 is w15, w14, ..., w0! */
    const __m128i shufmask = _mm_set_epi8 (0xff, 0xff, 3, 2,
                                           0xff, 0xff, 2, 1,
                                           0xff, 0xff, 1, 0,
                                           0xff, 0xff, 0xff, 0);

    const __m128i shufmask2 = _mm_set_epi8 (0xff, 0xff, 3, 2,
                                           0xff, 0xff, 3, 2,
                                           0xff, 0xff, 2, 1,
                                           0xff, 0xff, 1, 0);

    const __m128i mul = _mm_set_epi32 (1 << ((32 - 7) - 5),
                                       1 << ((32 - 7) - 6),
                                       1 << ((32 - 7) - 7),
                                       1 << ((32 - 7) - 0));

    const __m128i mul2 = _mm_set_epi32 (1 << ((32 - 7) - 2 - 7),
                                       1 << ((32 - 7) - 2),
                                       1 << ((32 - 7) - 3),
                                       1 << ((32 - 7) - 4));

    while (src < end)
    {
        /** load values 0-15 **/
        ba = _mm_loadu_si128 ((__m128i *) src);


        /** (1) decompress values 0-3 **/

        /* shuffle */
        ca = _mm_shuffle_epi8 (ba, shufmask);

        /* use multiplication (pmulld) to shift to left */
        ca = _mm_mullo_epi32 (ca, mul);

        /* use psrld to shift right */
        ca = _mm_srli_epi32 (ca, (32 -7));

        /* write out result */
        _mm_storeu_si128 (out, ca);
        out++;


        /** (2) decompress values 4-7 **/

        /* shift raw data by 4 bytes */
        ba = _mm_srli_si128 (ba, 3);

        /* shuffle */
        ca = _mm_shuffle_epi8 (ba, shufmask2);

        /* use multiplication (pmulld) to shift to left */
        ca = _mm_mullo_epi32 (ca, mul2);

        /* use psrld to shift right */
        ca = _mm_srli_epi32 (ca, (32 - 7));

        /* write out result */
        _mm_storeu_si128 (out, ca);
        out++;


        /** (3) decompress values 8-11 **/

        /* shift raw data by 5 bytes */
        ba = _mm_srli_si128 (ba, 4);

        /* shuffle */
        ca = _mm_shuffle_epi8 (ba, shufmask);

        /* use multiplication (pmulld) to shift to left */
        ca = _mm_mullo_epi32 (ca, mul);

        /* use psrld to shift right */
        ca = _mm_srli_epi32 (ca, (32 - 7));

        /* write out result */
        _mm_storeu_si128 (out, ca);
        out++;


        /** (4) decompress values 12-15 **/

        /* shift raw data by 4 bytes */
        ba = _mm_srli_si128 (ba, 3);

        /* shuffle */
        ca = _mm_shuffle_epi8 (ba, shufmask2);

        /* use multiplication (pmulld) to shift to left */
        ca = _mm_mullo_epi32 (ca, mul2);

        /* use psrld to shift right */
        ca = _mm_srli_epi32 (ca, (32 - 7));

        /* write out result */
        _mm_storeu_si128 (out, ca);
        out++;


        /** 16x7bit = 14 byte **/
        src += 14;
    }
}
#endif
#else
#if (SUMUP == 1)
static void
Serial_decompress7to32_sum(long unsigned *sum, uint8_t *src, unsigned long n)
{

    uint8_t *end = src + n * 7L / 8L;
    uint32_t temp;

    uint8_t * mask = malloc (8 * sizeof(uint8_t));

    mask[0] = 0x1;
    mask[1] = 0x3;
    mask[2] = 0x7;
    mask[3] = 0xf;
    mask[4] = 0x1f;
    mask[5] = 0x3f;
    mask[6] = 0x7f;
    mask[7] = 0xff;

    while (src < end) {

      *sum += src[0] & mask[6];

      temp  = src[0] >> 7;
      temp |= (src[1] & mask[5]) << 1;
      *sum += temp;

      temp  = src[1] >> 6;
      temp |= (src[2] & mask[4]) << 2;
      *sum += temp;

      temp  = src[2] >> 5;
      temp |= (src[3] & mask[3]) << 3;
      *sum += temp;

      temp  = src[3] >> 4;
      temp |= (src[4] & mask[2]) << 4;
      *sum += temp;

      temp  = src[4] >> 3;
      temp |= (src[5] & mask[1]) << 5;
      *sum += temp;

      temp  = src[5] >> 2;
      temp |= (src[6] & mask[0]) << 6;
      *sum += temp;

      *sum += src[6] >> 1;

      src  += 7;
    }

    free(mask);
}
#else
static void
Serial_decompress7to32(unsigned int *dest, uint8_t *src, unsigned long n)
{

    uint8_t *end = src + n * 7L / 8L;
    
    uint8_t * mask = malloc (8 * sizeof(uint8_t));

    mask[0] = 0x1;
    mask[1] = 0x3;
    mask[2] = 0x7;
    mask[3] = 0xf;
    mask[4] = 0x1f;
    mask[5] = 0x3f;
    mask[6] = 0x7f;
    mask[7] = 0xff;

    while (src < end) {

      dest[0]  = src[0] & mask[6];

      dest[1]  = src[0] >> 7;
      dest[1] |= (src[1] & mask[5]) << 1;

      dest[2]  = src[1] >> 6;
      dest[2] |= (src[2] & mask[4]) << 2; 

      dest[3]  = src[2] >> 5;
      dest[3] |= (src[3] & mask[3]) << 3; 

      dest[4]  = src[3] >> 4;
      dest[4] |= (src[4] & mask[2]) << 4; 

      dest[5]  = src[4] >> 3;
      dest[5] |= (src[5] & mask[1]) << 5; 

      dest[6]  = src[5] >> 2;
      dest[6] |= (src[6] & mask[0]) << 6; 

      dest[7]  = src[6] >> 1;

      dest += 8;
      src  += 7;
    }

    free(mask);
}
#endif
#endif

# elif ( MODE == 2 )
static uint8_t *
compress32to9 (unsigned int *data, unsigned long n)
{

    unsigned long m = (n * 9L/8L) + 1;
    uint8_t *ret = malloc (m * sizeof (*ret));

    long long offset = 0;
    for (long long i_in = 0; i_in < n; i_in+=8)
    {
      ret[offset+0]  = data[i_in+0];
      ret[offset+1]  = (data[i_in+0] >> 8) & 0x1;

      ret[offset+1] |= data[i_in+1] << 1;
      ret[offset+2] = (data[i_in+1] >> 7) & 0x3;
      
      ret[offset+2] |= data[i_in+2] << 2;
      ret[offset+3] = (data[i_in+2] >> 6) & 0x7;
      
      ret[offset+3] |= data[i_in+3] << 3;
      ret[offset+4] = (data[i_in+3] >> 5) & 0xf;
      
      ret[offset+4] |= data[i_in+4] << 4;
      ret[offset+5] = (data[i_in+4] >> 4) & 0x1f;
      
      ret[offset+5] |= data[i_in+5] << 5;
      ret[offset+6] = (data[i_in+5] >> 3) & 0x3f;

      ret[offset+6] |= data[i_in+6] << 6;
      ret[offset+7] = (data[i_in+6] >> 2) & 0x7f;

      ret[offset+7] |= (data[i_in+7] << 7);
      ret[offset+8] = (data[i_in+7] >> 1);

      offset += 9;
    }

    return ret;
}

#if( SIMD == 1 )
#if (SUMUP == 1)
static void
SIMD_decompress9to32_sum (long unsigned *sum, uint8_t *src, unsigned long n)
{
    uint8_t *end = src + n * 9 / 8;

    __m128i  ba;
    __m128i  ca;

    __m128i parallel_sum = _mm_setzero_si128();

     /* Note: Order of the arguments of _mm_set_epi8 is w15, w14, ..., w0! */
    const __m128i shufmask = _mm_set_epi8 (0xff, 0xff, 4, 3,
                                           0xff, 0xff, 3, 2,
                                           0xff, 0xff, 2, 1,
                                           0xff, 0xff, 1, 0);

    const __m128i mul = _mm_set_epi32 (1 << ((32 - 9) - 3),
                                       1 << ((32 - 9) - 2),
                                       1 << ((32 - 9) - 1),
                                       1 << ((32 - 9) - 0));

    while (src < end)
    {
        /** load values 0-11 **/
        ba = _mm_loadu_si128 ((__m128i *) src);

        
        /** decompress values 0-3 **/

        /* shuffle */
        ca = _mm_shuffle_epi8 (ba, shufmask);

        /* use multiplication (pmulld) to shift to left */
        ca = _mm_mullo_epi32 (ca, mul);

        /* use psrld to shift right */
        ca = _mm_srli_epi32 (ca, (32 -9));

        /* add up result */
        parallel_sum = _mm_add_epi32(parallel_sum, ca);


        /** decompress values 4-7 **/

        /* shift raw data by 4 bytes */
        ba = _mm_srli_si128 (ba, 4);

        /* shuffle */
        ca = _mm_shuffle_epi8 (ba, shufmask);

        /* shift left */
        ca = _mm_srli_epi32 (ca, 4);

        /* use multiplication (pmulld) to shift to left */
        ca = _mm_mullo_epi32 (ca, mul);

        /* use psrld to shift right */
        ca = _mm_srli_epi32 (ca, (32 - 9));

        /* add up result */
        parallel_sum = _mm_add_epi32(parallel_sum, ca);


        /** decompress values 8-11 **/

        /* shift raw data by 5 bytes */
        ba = _mm_srli_si128 (ba, 5);

        /* shuffle */
        ca = _mm_shuffle_epi8 (ba, shufmask);

        /* use multiplication (pmulld) to shift to left */
        ca = _mm_mullo_epi32 (ca, mul);

        /* use psrld to shift right */
        ca = _mm_srli_epi32 (ca, (32 - 9));

        /* add up result */
        parallel_sum = _mm_add_epi32(parallel_sum, ca);


        /** load values 12-23 **/
        src += 13;
        ba = _mm_loadu_si128 ((__m128i *) src);


        /** decompress values 12-15 **/

        /* shuffle */
        ca = _mm_shuffle_epi8 (ba, shufmask);

        /* shift left */
        ca = _mm_srli_epi32 (ca, 4);

        /* use multiplication (pmulld) to shift to left */
        ca = _mm_mullo_epi32 (ca, mul);

        /* use psrld to shift right */
        ca = _mm_srli_epi32 (ca, (32 - 9));

        /* add up result */
        parallel_sum = _mm_add_epi32(parallel_sum, ca);


        /** decompress values 16-19 **/

        /* shift raw data by 5 bytes */
        ba = _mm_srli_si128 (ba, 5);

        /* shuffle */
        ca = _mm_shuffle_epi8 (ba, shufmask);

        /* use multiplication (pmulld) to shift to left */
        ca = _mm_mullo_epi32 (ca, mul);

        /* use psrld to shift right */
        ca = _mm_srli_epi32 (ca, (32 - 9));

        /* add up result */
        parallel_sum = _mm_add_epi32(parallel_sum, ca);


        /** decompress values 20-23 **/

        /* shift raw data by 4 bytes */
        ba = _mm_srli_si128 (ba, 4);

        /* shuffle */
        ca = _mm_shuffle_epi8 (ba, shufmask);

        /* shift left */
        ca = _mm_srli_epi32 (ca, 4);

        /* use multiplication (pmulld) to shift to left */
        ca = _mm_mullo_epi32 (ca, mul);

        /* use psrld to shift right */
        ca = _mm_srli_epi32 (ca, (32 - 9));

        /* add up result */
        parallel_sum = _mm_add_epi32(parallel_sum, ca);


        src += 14;
    }

    *sum = _mm_extract_epi32(parallel_sum,0)+_mm_extract_epi32(parallel_sum,1)+_mm_extract_epi32(parallel_sum,2)+_mm_extract_epi32(parallel_sum,3);

}
#else
static void
SIMD_decompress9to32 (unsigned int *dest, uint8_t *src, unsigned long n)
{
    uint8_t *end = src + n * 9 / 8;

    __m128i *out = (__m128i *) dest;

    __m128i  ba;
    __m128i  ca;

     /* Note: Order of the arguments of _mm_set_epi8 is w15, w14, ..., w0! */
    const __m128i shufmask = _mm_set_epi8 (0xff, 0xff, 4, 3,
                                           0xff, 0xff, 3, 2,
                                           0xff, 0xff, 2, 1,
                                           0xff, 0xff, 1, 0);
    const __m128i mul = _mm_set_epi32 (1 << ((32 - 9) - 3),
                                       1 << ((32 - 9) - 2),
                                       1 << ((32 - 9) - 1),
                                       1 << ((32 - 9) - 0));

    while (src < end)
    {
        /** load values 0-11 **/
        ba = _mm_loadu_si128 ((__m128i *) src);


        /** decompress values 0-3 **/

        /* shuffle */
        ca = _mm_shuffle_epi8 (ba, shufmask);

        /* use multiplication (pmulld) to shift to left */
        ca = _mm_mullo_epi32 (ca, mul);

        /* use psrld to shift right */
        ca = _mm_srli_epi32 (ca, (32 -9));

        /* write out result */
        _mm_storeu_si128 (out, ca);
        out++;


        /** decompress values 4-7 **/

        /* shift raw data by 4 bytes */
        ba = _mm_srli_si128 (ba, 4);

        /* shuffle */
        ca = _mm_shuffle_epi8 (ba, shufmask);

        /* shift left */
        ca = _mm_srli_epi32 (ca, 4);

        /* use multiplication (pmulld) to shift to left */
        ca = _mm_mullo_epi32 (ca, mul);

        /* use psrld to shift right */
        ca = _mm_srli_epi32 (ca, (32 - 9));

        /* write out result */
        _mm_storeu_si128 (out, ca);
        out++;


        /** decompress values 8-11 **/

        /* shift raw data by 5 bytes */
        ba = _mm_srli_si128 (ba, 5);

        /* shuffle */
        ca = _mm_shuffle_epi8 (ba, shufmask);

        /* use multiplication (pmulld) to shift to left */
        ca = _mm_mullo_epi32 (ca, mul);

        /* use psrld to shift right */
        ca = _mm_srli_epi32 (ca, (32 - 9));

        /* write out result */
        _mm_storeu_si128 (out, ca);
        out++;


        /** load values 12-23 **/
        src += 13;
        ba = _mm_loadu_si128 ((__m128i *) src);


        /** decompress values 12-15 **/

        /* shuffle */
        ca = _mm_shuffle_epi8 (ba, shufmask);

        /* shift left */
        ca = _mm_srli_epi32 (ca, 4);

        /* use multiplication (pmulld) to shift to left */
        ca = _mm_mullo_epi32 (ca, mul);

        /* use psrld to shift right */
        ca = _mm_srli_epi32 (ca, (32 - 9));

        /* write out result */
        _mm_storeu_si128 (out, ca);
        out++;


        /** decompress values 16-19 **/

        /* shift raw data by 5 bytes */
        ba = _mm_srli_si128 (ba, 5);

        /* shuffle */
        ca = _mm_shuffle_epi8 (ba, shufmask);

        /* use multiplication (pmulld) to shift to left */
        ca = _mm_mullo_epi32 (ca, mul);

        /* use psrld to shift right */
        ca = _mm_srli_epi32 (ca, (32 - 9));

        /* write out result */
        _mm_storeu_si128 (out, ca);
        out++;


        /** decompress values 20-23 **/

        /* shift raw data by 4 bytes */
        ba = _mm_srli_si128 (ba, 4);

        /* shuffle */
        ca = _mm_shuffle_epi8 (ba, shufmask);

        /* shift left */
        ca = _mm_srli_epi32 (ca, 4);

        /* use multiplication (pmulld) to shift to left */
        ca = _mm_mullo_epi32 (ca, mul);

        /* use psrld to shift right */
        ca = _mm_srli_epi32 (ca, (32 - 9));

        /* write out result */
        _mm_storeu_si128 (out, ca);
        out++;

        src += 14;
    }
}
#endif
#else
#if (SUMUP == 1)
static void
Serial_decompress9to32_sum(long unsigned *sum, uint8_t *src, unsigned long n)
{

    uint8_t *end = src + n * 9L / 8L;
    
    uint8_t shift;
    uint8_t index;
    uint32_t temp;

    uint8_t * mask = malloc (8 * sizeof(uint8_t));

    mask[0] = 0x1;
    mask[1] = 0x3;
    mask[2] = 0x7;
    mask[3] = 0xf;
    mask[4] = 0x1f;
    mask[5] = 0x3f;
    mask[6] = 0x7f;
    mask[7] = 0xff;

    while (src < end) {
      shift = 0;
      for(index = 0; index < 8; index++) {
        temp = src[index] >> shift;
        temp |= (src[index+1] & (mask[shift])) << (8-shift);
        *sum += temp;
        shift++;
      }
      
      src  += 9;
    }

    free(mask);
}
#else
static void
Serial_decompress9to32 (unsigned int *dest, uint8_t *src, unsigned long n)
{

    uint8_t *end = src + n * 9L / 8L;
    
    uint8_t shift;
    uint8_t index;

    uint8_t * mask = malloc (8 * sizeof(uint8_t));

    mask[0] = 0x1;
    mask[1] = 0x3;
    mask[2] = 0x7;
    mask[3] = 0xf;
    mask[4] = 0x1f;
    mask[5] = 0x3f;
    mask[6] = 0x7f;
    mask[7] = 0xff;

    while (src < end) {
      shift = 0;
      for(index = 0; index < 8; index++) {
        dest[index]  = src[index] >> shift;
        dest[index] |= (src[index+1] & (mask[shift])) << (8-shift);
        shift++;
      }
      
      dest += 8;
      src  += 9;
    }

    free(mask);
}
#endif
#endif
#else

static uint8_t *
compress32to8 (unsigned int *data, unsigned long n)
{

    unsigned long m = n;
    uint8_t *ret = malloc (m * sizeof (*ret));

    long long offset = 0;
    for (long long i_in = 0; i_in < n; i_in++)
    {
      ret[offset]  = data[i_in] & 0xff;
      offset++;
    }

    return ret;
}

#if( SIMD == 1 )
#if (SUMUP == 1)
static void
SIMD_decompress8to32_sum(long unsigned *sum, uint8_t *src, unsigned long n)
{
    __m128i  ba;
    __m128i  ca;
    int i;

    uint8_t *end = src + n;
    __m128i parallel_sum = _mm_setzero_si128();;

     /* Note: Order of the arguments of _mm_set_epi8 is w15, w14, ..., w0! */
    const __m128i shufmask = _mm_set_epi8 (0xff, 0xff, 0xff, 3,
                                           0xff, 0xff, 0xff, 2,
                                           0xff, 0xff, 0xff, 1,
                                           0xff, 0xff, 0xff, 0);
    while (src < end)
    {
        /** load values 0-15 **/
        ba = _mm_loadu_si128 ((__m128i *) src);

        for(i=0; i<4; i++) {
          /* shuffle */
          ca = _mm_shuffle_epi8 (ba, shufmask);

          /* add up result */
          parallel_sum = _mm_add_epi32(parallel_sum, ca);

          /* shift raw data by 4 bytes */
          if(i<3) {
            ba = _mm_srli_si128 (ba, 4);
          }
        }

        src += 16;
    }

    *sum = _mm_extract_epi32(parallel_sum,0)+_mm_extract_epi32(parallel_sum,1)+_mm_extract_epi32(parallel_sum,2)+_mm_extract_epi32(parallel_sum,3);

}
#else
static void
SIMD_decompress8to32(unsigned int *dest, uint8_t *src, unsigned long n)
{
    __m128i  ba;
    __m128i  ca;
    int i;

    uint8_t *end = src + n;
    __m128i *out = (__m128i *) dest;

     /* Note: Order of the arguments of _mm_set_epi8 is w15, w14, ..., w0! */
    const __m128i shufmask = _mm_set_epi8 (0xff, 0xff, 0xff, 3,
                                           0xff, 0xff, 0xff, 2,
                                           0xff, 0xff, 0xff, 1,
                                           0xff, 0xff, 0xff, 0);

    while (src < end)
    {
        /** load values 0-15 **/
        ba = _mm_loadu_si128 ((__m128i *) src);

        for(i=0; i<4; i++) {
          /* shuffle */
          ca = _mm_shuffle_epi8 (ba, shufmask);

          /* write out result */
          _mm_storeu_si128 (out, ca);
          out++;

          /* shift raw data by 4 bytes */
          if(i<3) {
            ba = _mm_srli_si128 (ba, 4);
          }
        }

        src += 16;
    }
}
#endif
#else
#if (SUMUP == 1)
  static void
  Serial_decompress8to32_sum(long unsigned *sum, uint8_t *src, unsigned long n)
    {
      while (n--) {
      *sum += src[n];
    }
  }
#else
  static void
  Serial_decompress8to32(unsigned int *dest, uint8_t *src, unsigned long n)
    {
      while (n--) {
      dest[n] = src[n];
    }
  }
#endif
#endif
#endif

int
main (int argc, char **argv)
{
    const char    *progname = argv[0];
    unsigned long  n = N;
    unsigned int  *data;
    uint8_t       *compressed;
#if (SUMUP == 1)
  long unsigned sum = 0;
#else
  unsigned int *uncompressed;
#endif

    /* variables to measure execution times */
    long time1;
    struct timeval start, end;

    /* parse command line options */
    int ch;
    while ((ch = getopt (argc, argv, "hn:")) != -1)
    {
        switch (ch)
        {
            case 'h':
                usage (progname);
                break;

            case 'n':
                n = strtol (optarg, NULL, 0);
                break;
        }
    }

#if ( MODE == 1 )
    char* label = "32-7";
#elif ( MODE == 2 )
     char* label = "32-9";
#else
     char* label = "32-8";
#endif

    /* generate random data */
    printf ("Generating data %s ...", label);
    data = generate (n);
    printf ("done.\n");

    /* compress data */
    printf ("Compressing data...");
#if ( MODE == 1 )
    compressed = compress32to7 (data, n);
#elif ( MODE == 2 )
    compressed = compress32to9 (data, n);
#else
    compressed = compress32to8 (data, n);
#endif
    printf ("done.\n");

#if ( SUMUP == 0 )
    printf ("Allocating memory for decompression result...");
    uncompressed = malloc ((n + 4) * sizeof (*uncompressed));
    if (!uncompressed)
    {
        fprintf (stderr, "out of memory\n");
        exit (EXIT_FAILURE);
    }
    memset (uncompressed, 0, n * sizeof (*uncompressed));
    printf ("done.\n");
#endif

#if( SIMD == 1 )
  #if( SUMUP == 1 )
    printf("-> Decompressing + Accumulating (vectorized)...");
  #else
    printf("-> Decompressing (vectorized)...");
  #endif
#else
  #if( SUMUP == 1 )
    printf("-> Decompressing + Accumulating (serially)...");
  #else
    printf("-> Decompressing (serially)...");
  #endif
#endif

    /* remember start time */
    gettimeofday (&start, NULL);
    
#if ( MODE == 1 )
  #if ( SIMD == 1 )
    #if ( SUMUP == 0 )
      SIMD_decompress7to32(uncompressed, compressed, n);
    #else
      SIMD_decompress7to32_sum(&sum, compressed, n);
    #endif
  #else
    #if ( SUMUP == 0 )
      Serial_decompress7to32(uncompressed, compressed, n);
    #else
      Serial_decompress7to32_sum(&sum, compressed, n);
    #endif
  #endif
#elif ( MODE == 2 )
  #if ( SIMD == 1 )
    #if ( SUMUP == 0 )
      SIMD_decompress9to32(uncompressed, compressed, n);
    #else
      SIMD_decompress9to32_sum(&sum, compressed, n);
    #endif
  #else
    #if ( SUMUP == 0 )
      Serial_decompress9to32(uncompressed, compressed, n);
    #else
      Serial_decompress9to32_sum(&sum, compressed, n);
    #endif
  #endif
#else
  #if ( SIMD == 1 )
    #if ( SUMUP == 0 )
      SIMD_decompress8to32(uncompressed, compressed, n);
    #else
      SIMD_decompress8to32_sum(&sum, compressed, n);
    #endif
  #else
    #if ( SUMUP == 0 )
       Serial_decompress8to32(uncompressed, compressed, n);
    #else
       Serial_decompress8to32_sum(&sum, compressed, n);
    #endif
  #endif
#endif

    /* get end time... */
    gettimeofday (&end, NULL);

    /* ...and compute the difference */
    time1 = (end.tv_sec - start.tv_sec) * 1000000
        + (end.tv_usec - start.tv_usec);

    printf ("done.\n");

    printf ("Required %lu usec to decompress.\n", time1);

#if ( SUMUP == 0 )
    /* verify */
    bool verify = true;
    unsigned long i;
    for (i = 0; i < n; i++)
    {
        if (data[i] != uncompressed[i])   {
            verify = false;
            break;
        }
    }

    if(verify)
      printf ("comparison: successful\n");
    else
      printf ("comparison failed at position %d : %d != %d\n", (int) i, data[i], (uint8_t) uncompressed[i]);

    printf ("\n");
    free(uncompressed);
#else
    /* verify */
    unsigned long i;
    long unsigned actual_sum = 0;
    for (i = 0; i < n; i++)
    {
        actual_sum += data[i];
    }

    if(actual_sum == sum)
      printf ("sum: successful\n");
    else
      printf ("sum failed %ld != %ld\n", sum, actual_sum);

    printf ("\n");
#endif

  // clean up
  free(data);
  free(compressed);

}
